/*
 * Copyright 2005 Novell, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that the above copyright notice appear in all copies
 * and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of
 * Novell, Inc. not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission.
 * Novell, Inc. makes no representations about the suitability of this
 * software for any purpose. It is provided "as is" without express or
 * implied warranty.
 *
 * NOVELL, INC. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN
 * NO EVENT SHALL NOVELL, INC. BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
 * OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: David Reveman <davidr@novell.com>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <sys/time.h>
#include <GL/glut.h>

#include "debug.h"
#include "debug_uglyfont.h"
#include "bang.h"
#include "bang_window.h"



/**********************************
 *
 * E_Action callbacks
 *
 **********************************/
void
_debug_action_cb(E_Object *obj, const char *params)
{
   DD(("_debug_action_cb\n"));
}

void
_debug_action_mouse_cb(E_Object *obj, const char *params, Ecore_X_Event_Mouse_Button_Down *ev)
{
   DD(("_debug_action_mouse_cb\n"));
}

void
_debug_action_key_cb(E_Object *obj, const char *params, Ecore_X_Event_Key_Down *ev)
{
   DD(("_debug_action_key_cb\n"));
   debug_console_full();
   debug->active = !debug->active;
   
   
   //TODO Fix this way to retrive screen
   CompDisplay *d = bang->displays->data;
   CompScreen  *s;
   s = findScreenAtDisplay(d,ecore_x_window_root_first_get());
   
   damageScreen(s);
   
}

/**********************************
 *
 * Wrapped functions
 *
 **********************************/

static Bool
debug_paint_window(CompWindow              *w,
                   const WindowPaintAttrib *attrib,
                   Region                   region,
                   unsigned int             mask)
{
   float r, g, b, a;
   if (!w->mapNum)
      return;
   CompScreen *s;
   Bool status;
   s = w->screen;
   
   
   
   //UNWRAP (debug, s, paintWindow);
   //status = (*s->paintWindow) (w, attrib, region, mask);
   //WRAP (debug, s, paintWindow, debug_paint_window);
   status = (*debug->paintWindow) (w, attrib, region, mask);
   
   //~ if (!debug || !debug->active || !region->numRects) return status;
   
   //~ DD(("debug_paint_window [x: %d y: %d]\n", w->serverX, w->serverY));
   //~ DD(("debug_paint_window [x0: %f y0: %f]\n", w->matrix.x0, w->matrix.y0));
   
   //~ glPushMatrix ();
   //~ glLoadIdentity();

   //~ r = 0.0f;
   //~ g = 0.1f;
   //~ b = 0.0f;
   //~ a = 0.5f;

   //~ glEnable (GL_BLEND);
   //~ glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
   
   //~ // Rect
   //~ glBegin(GL_POLYGON);
   //~ glColor4f(1.0f, 0.0f, 0.0f, a);
   //~ glVertex3f(-1.0f, -0.5f, -2.0f);   // D
  //~ // glVertex3f(w->attrib.x, -w->attrib.y, -0.0f);   // D
   
   //~ glColor4f(0.0f, 1.0f, 0.0f, a );
   //~ glVertex3f(1.0f, -0.5f, -2.0f);    // A
   //~ glColor4f(0.0f, 0.0f, 1.0f, a);
   //~ glVertex3f( 1.0f, 0.5f, -2.0f);    // B
   //~ glVertex3f( -1.0f,  0.5f, -2.0f);  // C
  //~ // glVertex3f( w->matrix.x0,  w->matrix.y0, -2.0f);   // D
   //~ glEnd ();
   
   //~ glPopMatrix();
   //~ glColor4usv(defaultColor); //Remove this line for a special effect
   
   //~ damageScreen(s);
   return status;
}



static Bool
debug_paint_screen(CompScreen              *s,
                   const ScreenPaintAttrib *sAttrib,
                   Region                   region,
                   unsigned int             mask)
{
   Bool status;
   float r, g, b, a;
   
   /* Call original paintScreen function */
   UNWRAP (debug, s, paintScreen);
   status = (*s->paintScreen) (s, sAttrib, region, mask);
   WRAP (debug, s, paintScreen, debug_paint_screen);

   
   if (!debug || !debug->active || !region->numRects) return status;
   DD(("debug_paint_screen \n"));
   /* Draw main info box */
   glPushMatrix ();

   r = 0.0f;
   g = 0.1f;
   b = 0.0f;
   a = 0.5f;

   glEnable (GL_BLEND);
   glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
   
   // Rect
   glBegin(GL_POLYGON);
   glColor4f(1.0f, 0.0f, 0.0f, a);
   glVertex3f(-1.0f, -0.5f, -4.0f);   // D
   glColor4f(0.0f, 1.0f, 0.0f, a );
   glVertex3f(1.0f, -0.5f, -4.0f);    // A
   glColor4f(0.0f, 0.0f, 1.0f, a);
   glVertex3f( 1.0f, 0.5f, -4.0f);    // B
   glVertex3f( -1.0f,  0.5f, -4.0f);  // C
   glEnd ();
   
   // Border
   glColor4f(r, g, b, 1.0f);
   glBegin(GL_LINE_LOOP);
   glVertex3f(-1.0f, 0.5f, -4.0f);    // A
   glVertex3f( 1.0f, 0.5f, -4.0f);    // B
   glVertex3f( 1.0f,  -0.5f, -4.0f);  // C
   glVertex3f( -1.0f,  -0.5f, -4.0f); // D
   glEnd ();
  
   
   
   //Text
   #define SIZE 0.05
   glTranslatef(-0.9f,0.4f,-4.0f);
   
   glScalef(SIZE/2,SIZE,0.0f);
   glRotatef(45.0f,0.0f,0.0f,1.0f);
   glColor4f(0.0f, 0.0f, 0.0f, 1.0f);
   YsDrawUglyFont("Hello OpenGL World !!!",0,1);
   
   
   
   glPopMatrix();
   glColor4usv(defaultColor); //Remove this line for a special effect
   
   
   return status;
}

static void
debug_prepare_paint_screen(CompScreen *s,
                           int         msSinceLastPaint)
{
   DD(("debug_prepare_paint_screen \n"));

   //~ UNWRAP (ss, s, preparePaintScreen);
   //~ (*s->preparePaintScreen) (s, msSinceLastPaint);
   //~ WRAP (ss, s, preparePaintScreen, scalePreparePaintScreen);
}

static void
debug_done_paint_screen(CompScreen *s)
{
   DD(("debug_done_paint_screen \n"));

   //~ UNWRAP (ss, s, donePaintScreen);
   //~ (*s->donePaintScreen) (s);
   //~ WRAP (ss, s, donePaintScreen, scaleDonePaintScreen);
}


/**********************************
 *
 * Plugin functions
 *
 **********************************/
static Bool
debug_init_display(CompPlugin *p, CompDisplay *d)
{
   DD(("debug_init_display\n"));
  // ScaleDisplay *sd;

   //~ sd = malloc (sizeof (ScaleDisplay));
   //~ if (!sd)
      //~ return FALSE;

   //~ sd->screenPrivateIndex = allocateScreenPrivateIndex (d);
   //~ if (sd->screenPrivateIndex < 0)
   //~ {
      //~ free (sd);
      //~ return FALSE;
   //~ }

   //~ sd->lastActiveNum = None;

//~ ////    WRAP (sd, d, handleEvent, scaleHandleEvent);

   //~ d->privates[displayPrivateIndex].ptr = sd;

  // handle_key = ecore_event_handler_add(ECORE_X_EVENT_KEY_DOWN, _e_scale_cb_key_down, d);
  // handle_border_add = ecore_event_handler_add(E_EVENT_BORDER_ADD, _e_scale_cb_border_add, d);
  // handle_border_remove = ecore_event_handler_add(E_EVENT_BORDER_REMOVE, _e_scale_cb_border_remove, d);
   return TRUE;
}

static void
debug_fini_display(CompPlugin  *p, CompDisplay *d)
{
   DD(("debug_fini_display\n"));
}

static Bool
debug_init_screen(CompPlugin *p, CompScreen *s)
{
   DD(("debug_init_screen\n"));

   //~ WRAP (ss, s, preparePaintScreen, scalePreparePaintScreen);
   //~ WRAP (ss, s, donePaintScreen, scaleDonePaintScreen);
   WRAP (debug, s, paintScreen, debug_paint_screen);
   WRAP (debug, s, paintWindow, debug_paint_window);
   //~ WRAP (ss, s, damageWindowRect, scaleDamageWindowRect);

    return TRUE;
}

static void
debug_fini_screen(CompPlugin *p, CompScreen *s)
{
   DD(("debug_fini_screen\n"));

   //~ UNWRAP (ss, s, preparePaintScreen);
   //~ UNWRAP (ss, s, donePaintScreen);
   //~ UNWRAP (ss, s, paintScreen);
   //~ UNWRAP (ss, s, paintWindow);
   //~ UNWRAP (ss, s, damageWindowRect);

}

static Bool
debug_init_window(CompPlugin *p, CompWindow *w)
{
   DD(("debug_init_window\n"));

   return TRUE;
}

static void
debug_fini_window(CompPlugin *p, CompWindow *w)
{
   DD(("debug_fini_window\n"));
}

static Bool
debug_init(CompPlugin *p)
{
   DD(("debug_init\n"));

   
   debug = E_NEW(Debug, 1);
   if (!debug) return FALSE;
   
   debug->active = FALSE;
   
   /* Add a new e17 action and attach functions to it */
   /* The user can then bind the action to a key/mouse events */
#if DEBUG
   E_Action *action;
   action = e_action_add("bang_debug_on_console");
   if (action)
   {
      action->func.go = _debug_action_cb;
      action->func.go_mouse = _debug_action_mouse_cb;
      action->func.go_key = _debug_action_key_cb;
      e_action_predef_name_set("Bang", "Plugin debug on console",
                               "bang_debug_on_console", NULL, NULL, 0);
   }
#endif
   return TRUE;
}

static void
debug_fini(CompPlugin *p)
{
   DD(("debug_fini\n"));
   
   if (!debug) return;

#if DEBUG
   /* Remove e17 action */
   e_action_predef_name_del("Bang", "Plugin debug on console");
   e_action_del("bang_debug_on_console");
#endif
}


/**********************************
 *
 * Plugin vTable
 *
 **********************************/

/* Describe here your plugin.*/
/* All the functions will be called by the core when events occur. */
/* You can set a function handler to NULL if you are not intrested */
CompPluginVTable debugVTable = {
   "debug",               /* plugin name */
   "Debug Plugin",        /* plugin short description (and long descr follow)*/
   "Debug is a plugin used to debug bang module and is also a template for new plugins.",
   debug_init,            /* the function to call to init the plugin */
   debug_fini,            /* the function to call to destroy the plugin */
   debug_init_display,    /* this is called for every display created */
   debug_fini_display,    /* Destroy display */
   debug_init_screen,     /* this is called for every screen created */
   debug_fini_screen,     /* Destroy screen */
   debug_init_window,     /* this is called for every window created */
   debug_fini_window      /* Destroy windows */
};

/* This function is called by bang to obtain the plugin information */
/* Is the starting point of this plugin. */
CompPluginVTable *
getCompPluginInfo (void)
{
    return &debugVTable;
}
